use bitvec::index::BitIdx;
use std::{error::Error, str::FromStr, vec::Vec};

use deku::{
    bitvec::{BitStore, Msb0},
    ctx::ByteSize,
    deku_derive,
    reader::{Reader, ReaderRet},
    DekuError, DekuReader,
};

#[deku_derive(DekuRead)]
#[derive(Debug, Clone, Copy, PartialEq)]
#[deku(endian = "big")]
pub struct GroupId(u16);

trait MapToDekuParseError<T> {
    fn map_to_deku_parse_error(self) -> Result<T, DekuError>;
}

impl<T, E: Error> MapToDekuParseError<T> for Result<T, E> {
    fn map_to_deku_parse_error(self) -> Result<T, DekuError> {
        self.map_err(|e| DekuError::Parse(e.to_string().into()))
    }
}

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct DataField {
    #[deku(reader = "read_size(deku::reader)")]
    data_size: ByteSize,

    #[deku(bytes_read = "data_size.0")]
    pub data: Vec<u8>,
}

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct BlockHeader {
    pub group_id: GroupId,

    #[deku(reader = "read_size(deku::reader)")]
    pub data_size: ByteSize,
}

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct DataBlock {
    pub header: BlockHeader,

    #[deku(ctx = "header.group_id")]
    pub inner: DataGroup,
}

fn read_size<R: std::io::Read>(reader: &mut Reader<R>) -> Result<ByteSize, DekuError> {
    let first_byte: u8 = u8::from_reader_with_ctx(reader, ())?;

    let is_length_expanded = first_byte.get_bit::<Msb0>(BitIdx::new(0).map_to_deku_parse_error()?);

    match is_length_expanded {
        true => {
            let size_of_data_size: ByteSize = ByteSize((first_byte & 0b0111_1111) as usize);

            if size_of_data_size.0 > 4 {
                return Err(DekuError::Parse("Size of the length encoding is > 4, this is not normal. Probable parsing error".to_string().into()));
            };

            // maximum size of the buffer is 4, we use the offset to read values less than 4 bytes
            let buffer: &mut [u8; 4] = &mut [0; 4];
            let write_offset = 4 - size_of_data_size.0;

            match reader.read_bytes(size_of_data_size.0, &mut buffer[write_offset..])? {
                ReaderRet::Bits(_bit_vec) => Err(DekuError::Parse("Got bits when trying to read bytes -> reader is unaligned, this is not normal.".to_string().into())),
                ReaderRet::Bytes => Ok(ByteSize(u32::from_be_bytes(*buffer) as usize)),
            }
        }
        false => Ok(ByteSize(first_byte as usize)),
    }
}

// Using this as the map function asks deku to parse a datafield
// We then use the datafield and convert it to the corresponding value
fn convert_from_data_field<T>(data_field: DataField) -> Result<T, DekuError>
where
    T: FromStr,
    T::Err: Error,
{
    let text = String::from_utf8(data_field.data).map_to_deku_parse_error()?;
    T::from_str(&text).map_to_deku_parse_error()
}

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct SSVVersionNumber(#[deku(map = "convert_from_data_field")] u16);

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct GALSSVersionNumber(#[deku(map = "convert_from_data_field")] u16);

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct PSSVersionNumber(#[deku(map = "convert_from_data_field")] u16);

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct ConfigurationHeader {
    pub ssv_version: SSVVersionNumber,
    pub galss_version: GALSSVersionNumber,
    pub pss_version: PSSVersionNumber,
}

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct PCSCReaderName(#[deku(map = "convert_from_data_field")] String);

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct CardType(#[deku(map = "convert_from_data_field")] u8);

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct PCSCReader {
    pub name: PCSCReaderName,
    pub card_type: CardType,
}

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct SESAMVitaleComponentID(#[deku(map = "convert_from_data_field")] u16);

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct SESAMVitaleComponentDescription(#[deku(map = "convert_from_data_field")] String);

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct SESAMVitaleComponentVersion(#[deku(map = "convert_from_data_field")] String);

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct SESAMVitaleComponent {
    pub id: SESAMVitaleComponentID,
    pub description: SESAMVitaleComponentDescription,
    pub version: SESAMVitaleComponentVersion,
}

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
#[deku(ctx = "group_id: GroupId", id = "group_id.0")]
pub enum DataGroup {
    #[deku(id = 60)]
    ConfigurationHeader(ConfigurationHeader),
    #[deku(id = 67)]
    PCSCReader(PCSCReader),
    #[deku(id = 64)]
    SESAMVitaleComponent(SESAMVitaleComponent),
}
